/*
/ MapView.cpp
/ a panel to show the Map of LibreAtlas
/
/ version 1.0, 2012 May 16
/
/ Author: Sandro Furieri a-furieri@lqt.it
/
/ Copyright (C) 2012  Alessandro Furieri
/
/    This program is free software: you can redistribute it and/or modify
/    it under the terms of the GNU General Public License as published by
/    the Free Software Foundation, either version 3 of the License, or
/    (at your option) any later version.
/
/    This program is distributed in the hope that it will be useful,
/    but WITHOUT ANY WARRANTY; without even the implied warranty of
/    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/    GNU General Public License for more details.
/
/    You should have received a copy of the GNU General Public License
/    along with this program.  If not, see <http://www.gnu.org/licenses/>.
/
*/

#include "Classdef.h"
#include "wx/mstream.h"

#include <float.h>

#include "icons/crosshair.xpm"
#include "icons/pin.xpm"

MyMapView::MyMapView(MyFrame * parent, wxWindowID id):
wxPanel(parent, id, wxDefaultPosition, wxSize(440, 480),
        wxBORDER_SUNKEN | wxWANTS_CHARS)
{
//
// constructor: a frame to show the Map
//
  MainFrame = parent;
  CurrentGeometry = NULL;
  DistancePath = NULL;
  TimerMouseWheel = NULL;
  TimerMouseMove = NULL;
  TimerDynamic = NULL;
  MapTip = NULL;
  Invalidate();
  Pin = wxBitmap(pin_xpm);

  BitmapWidth = wxSystemSettings::GetMetric(wxSYS_SCREEN_X);
  BitmapHeight = wxSystemSettings::GetMetric(wxSYS_SCREEN_Y);
  MapBitmap = wxBitmap(BitmapWidth, BitmapHeight, -1);
  DynamicMapBitmap = wxBitmap(BitmapWidth, BitmapHeight, -1);
  ScreenBitmap = wxBitmap(BitmapWidth, BitmapHeight, -1);

  wxImage crossImage = wxBitmap(crosshair_xpm).ConvertToImage();
#ifdef __WXMSW__
// setting up a WINDOWS cursor
  crossImage.SetOption(wxIMAGE_OPTION_CUR_HOTSPOT_X, 8);
  crossImage.SetOption(wxIMAGE_OPTION_CUR_HOTSPOT_Y, 8);
  CursorCross = wxCursor(crossImage);
#else
#ifdef __WXMAC__
// setting up a MacOsX cursor
  crossImage.SetOption(wxIMAGE_OPTION_CUR_HOTSPOT_X, 8);
  crossImage.SetOption(wxIMAGE_OPTION_CUR_HOTSPOT_Y, 8);
  CursorCross = wxCursor(crossImage);
#else
// setting up an UNIX -likecursor
  int row;
  int col;
  unsigned short bits[16];
  unsigned short mask[16];
  for (row = 0; row < 16; row++)
    {
      bits[row] = 0x0000;
      mask[row] = 0x0000;
      for (col = 0; col < 16; col++)
        {
          unsigned char pixel = crossImage.GetRed(col, row);
          if (pixel == 0)
            {
              switch (col)
                {
                  case 0:
                    bits[row] |= 0x8000;
                    break;
                  case 1:
                    bits[row] |= 0x4000;
                    break;
                  case 2:
                    bits[row] |= 0x2000;
                    break;
                  case 3:
                    bits[row] |= 0x1000;
                    break;
                  case 4:
                    bits[row] |= 0x0800;
                    break;
                  case 5:
                    bits[row] |= 0x0400;
                    break;
                  case 6:
                    bits[row] |= 0x0200;
                    break;
                  case 7:
                    bits[row] |= 0x0100;
                    break;
                  case 8:
                    bits[row] |= 0x0080;
                    break;
                  case 9:
                    bits[row] |= 0x0040;
                    break;
                  case 10:
                    bits[row] |= 0x0020;
                    break;
                  case 11:
                    bits[row] |= 0x0010;
                    break;
                  case 12:
                    bits[row] |= 0x0008;
                    break;
                  case 13:
                    bits[row] |= 0x0004;
                    break;
                  case 14:
                    bits[row] |= 0x0002;
                    break;
                  case 15:
                    bits[row] |= 0x0001;
                    break;
                };
              switch (col)
                {
                  case 0:
                    mask[row] |= 0x8000;
                    break;
                  case 1:
                    mask[row] |= 0x4000;
                    break;
                  case 2:
                    mask[row] |= 0x2000;
                    break;
                  case 3:
                    mask[row] |= 0x1000;
                    break;
                  case 4:
                    mask[row] |= 0x0800;
                    break;
                  case 5:
                    mask[row] |= 0x0400;
                    break;
                  case 6:
                    mask[row] |= 0x0200;
                    break;
                  case 7:
                    mask[row] |= 0x0100;
                    break;
                  case 8:
                    mask[row] |= 0x0080;
                    break;
                  case 9:
                    mask[row] |= 0x0040;
                    break;
                  case 10:
                    mask[row] |= 0x0020;
                    break;
                  case 11:
                    mask[row] |= 0x0010;
                    break;
                  case 12:
                    mask[row] |= 0x0008;
                    break;
                  case 13:
                    mask[row] |= 0x0004;
                    break;
                  case 14:
                    mask[row] |= 0x0002;
                    break;
                  case 15:
                    mask[row] |= 0x0001;
                    break;
                };
            }
          if (pixel == 255)
            {
              switch (col)
                {
                  case 0:
                    mask[row] |= 0x8000;
                    break;
                  case 1:
                    mask[row] |= 0x4000;
                    break;
                  case 2:
                    mask[row] |= 0x2000;
                    break;
                  case 3:
                    mask[row] |= 0x1000;
                    break;
                  case 4:
                    mask[row] |= 0x0800;
                    break;
                  case 5:
                    mask[row] |= 0x0400;
                    break;
                  case 6:
                    mask[row] |= 0x0200;
                    break;
                  case 7:
                    mask[row] |= 0x0100;
                    break;
                  case 8:
                    mask[row] |= 0x0080;
                    break;
                  case 9:
                    mask[row] |= 0x0040;
                    break;
                  case 10:
                    mask[row] |= 0x0020;
                    break;
                  case 11:
                    mask[row] |= 0x0010;
                    break;
                  case 12:
                    mask[row] |= 0x0008;
                    break;
                  case 13:
                    mask[row] |= 0x0004;
                    break;
                  case 14:
                    mask[row] |= 0x0002;
                    break;
                  case 15:
                    mask[row] |= 0x0001;
                    break;
                };
            }
        }
    }
  CursorCross =
    wxCursor((char *) bits, 16, 16, 8, 8, (char *) mask, wxWHITE, wxBLACK);
#endif
#endif
  CursorHand = wxCursor(wxCURSOR_HAND);
  SetCursor(CursorCross);

// setting up event handlers
  Connect(wxID_ANY, wxEVT_ERASE_BACKGROUND,
          (wxObjectEventFunction) & MyMapView::OnEraseBackground);
  Connect(wxID_ANY, wxEVT_SIZE, (wxObjectEventFunction) & MyMapView::OnSize);
  Connect(wxID_ANY, wxEVT_PAINT, (wxObjectEventFunction) & MyMapView::OnPaint);
  Connect(wxID_ANY, wxEVT_MOTION, wxMouseEventHandler(MyMapView::OnMouseMove),
          NULL, this);
  Connect(ID_MOUSE_TIMER, wxEVT_TIMER,
          wxTimerEventHandler(MyMapView::OnTimerMouseMove), NULL, this);
  Connect(wxID_ANY, wxEVT_LEFT_DOWN,
          wxMouseEventHandler(MyMapView::OnMouseClick), NULL, this);
  Connect(wxID_ANY, wxEVT_RIGHT_DOWN,
          wxMouseEventHandler(MyMapView::OnContextMenu), NULL, this);
  Connect(wxID_ANY, wxEVT_MOUSEWHEEL,
          wxMouseEventHandler(MyMapView::OnMouseWheel), NULL, this);
  Connect(ID_WHEEL_TIMER, wxEVT_TIMER,
          wxTimerEventHandler(MyMapView::OnTimerMouseWheel), NULL, this);
  Connect(wxID_ANY, wxEVT_LEFT_UP,
          wxMouseEventHandler(MyMapView::OnMouseDragStop), NULL, this);
  Connect(wxID_ANY, wxEVT_LEAVE_WINDOW,
          wxMouseEventHandler(MyMapView::OnMouseDragStop), NULL, this);
  Connect(ID_DYNAMIC_TIMER, wxEVT_TIMER,
          wxTimerEventHandler(MyMapView::OnTimerDynamic), NULL, this);
}

MyMapView::~MyMapView()
{
  if (TimerMouseMove)
    {
      TimerMouseMove->Stop();
      delete TimerMouseMove;
    }
  if (TimerMouseWheel)
    {
      TimerMouseWheel->Stop();
      delete TimerMouseWheel;
    }
  if (TimerDynamic)
    {
      TimerDynamic->Stop();
      delete TimerDynamic;
      TimerDynamic = NULL;
    }
  ClearMapTip();
  ResetCurrentGeometry();
  if (DistancePath != NULL)
    gaiaFreeGeomColl(DistancePath);
}

void MyMapView::Invalidate()
{
// resetting to initial state
  ValidMap = false;
  PixelRatio = DBL_MAX;
  CurrentScale = 0;
  DragStartX = -1;
  DragStartY = -1;
  LastDragX = -1;
  LastDragY = -1;
  DistanceX1 = DBL_MAX;
  DistanceY1 = DBL_MAX;
  DistanceX2 = DBL_MAX;
  DistanceY2 = DBL_MAX;
  if (TimerMouseMove)
    {
      TimerMouseMove->Stop();
      delete TimerMouseMove;
    }
  TimerMouseMove = NULL;
  if (TimerMouseWheel)
    {
      TimerMouseWheel->Stop();
      delete TimerMouseWheel;
    }
  TimerMouseWheel = NULL;
  if (TimerDynamic)
    {
      TimerDynamic->Stop();
      delete TimerDynamic;
      TimerDynamic = NULL;
    }
  TimerDynamic = NULL;
  WheelTics = 0;
  if (MapTip)
    ClearMapTip();
  MapTip = NULL;
  ResetCurrentGeometry();
  if (DistancePath)
    gaiaFreeGeomColl(DistancePath);
  DistancePath = NULL;
}

void MyMapView::SetCurrentGeometry(gaiaGeomCollPtr geom)
{
// setting the current Geometry
  if (CurrentGeometry != NULL)
    gaiaFreeGeomColl(CurrentGeometry);
  CurrentGeometry = geom;
}

void MyMapView::ResetCurrentGeometry()
{
// resetting the current Geometry
  if (CurrentGeometry != NULL)
    gaiaFreeGeomColl(CurrentGeometry);
  CurrentGeometry = NULL;
}

void MyMapView::OnSize(wxSizeEvent & WXUNUSED(event))
{
//
// this window has changed its size
//

// refreshing the map
  PrepareMap();
  ClearMapTip();
}

void MyMapView::OnPaint(wxPaintEvent & WXUNUSED(event))
{
//
// this window needs to be repainted
//
  wxPaintDC dc(this);
  if (ValidMap == false)
    {
      // no map: black background
      wxSize sz = GetClientSize();
      dc.SetBrush(wxBrush(wxColour(0, 0, 0), wxSOLID));
      dc.SetPen(wxPen(wxColour(0, 0, 0), 1));
      dc.DrawRectangle(0, 0, sz.GetWidth(), sz.GetHeight());
      dc.SetBrush(wxNullBrush);
      dc.SetPen(wxNullPen);
  } else if (dc.IsOk() && ScreenBitmap.IsOk())
    {
      wxMemoryDC *memDC = new wxMemoryDC(ScreenBitmap);
      dc.Blit(0, 0, BitmapWidth, BitmapHeight, memDC, 0, 0);
      delete memDC;
    }
  MainFrame->UpdateStatusBar();
}

void MyMapView::OnMouseMove(wxMouseEvent & event)
{
//
// MOUSE motion - event handler
// 
  double x;
  double y;
  char dummy[1024];
  wxString coords;
  if (MapTip)
    {
// check if the current Map Tip has to be erased
      if (MapTip->IsToBeErased(event.GetX(), event.GetY()) == true)
        ClearMapTip();
    }
  if (event.LeftIsDown() == true && MainFrame->IsModePan() == true)
    {
      SetCursor(CursorHand);
      if (abs(LastDragX - event.GetX()) > 25
          || abs(LastDragY - event.GetY()) > 25)
        DragMap(event.GetX(), event.GetY());
      return;
    }
  SetCursor(CursorCross);
  x = FrameMinX + ((double) (event.GetX()) * PixelRatio);
  y = FrameMaxY - ((double) (event.GetY()) * PixelRatio);
  if (x >= MainFrame->GetMapMinX() && x <= MainFrame->GetMapMaxX()
      && y >= MainFrame->GetMapMinY() && y <= MainFrame->GetMapMaxY())
    ;
  else
    {
      coords = wxT("");
      MainFrame->UpdateCoords(coords);
      return;
    }

  if (MainFrame->IsGeographic() == true)
    {
      // geographic coordinates DMS
      char north = 'N';
      char east = 'E';
      double absLong = x;
      double absLat = y;
      double longDegs;
      double longMins;
      double longSecs;
      double latDegs;
      double latMins;
      double latSecs;
      double fract;
      double fract2;
      if (x < 0.0)
        {
          east = 'W';
          absLong = x * -1.0;
        }
      if (y < 0.0)
        {
          north = 'S';
          absLat = y * -1.0;
        }
      longDegs = floor(absLong);
      fract = (absLong - longDegs) * 60.0;
      longMins = floor(fract);
      fract2 = (fract - longMins) * 60.0;
      longSecs = floor(fract2);

      latDegs = floor(absLat);
      fract = (absLat - latDegs) * 60.0;
      latMins = floor(fract);
      fract2 = (fract - latMins) * 60.0;
      latSecs = floor(fract2);
      sprintf(dummy,
              "Lat: %3.0f^%02.0f'%02.0f\"%c         Lon: %3.0f^%02.0f'%02.0f\"%c",
              latDegs, latMins, latSecs, north, longDegs, longMins, longSecs,
              east);
      wxString dms = wxString::FromUTF8(dummy);
      wxString degree(wxT("?"));
      degree.SetChar(0, 0xB0);
      dms.Replace(wxT("^"), degree);
      coords = dms;

      // current scale (Geographic Coords)
      double cm = PixelRatio * 72.0 * 0.3937;
      CurrentScale = (int) (cm * 11113200.0);
  } else
    {
      // planar coordinates
      sprintf(dummy, "X: %1.6f         Y: %1.6f", x, y);
      coords = wxString::FromUTF8(dummy);
      double cm = PixelRatio * 72.0 * 0.3937;
      CurrentScale = (int) (cm * 100.0);
    }
  MainFrame->UpdateMapInfo();
  MainFrame->UpdateCoords(coords);
  MainFrame->UpdateStatusBar();

  // starting the timer
  if (!TimerMouseMove)
    TimerMouseMove = new wxTimer(this, ID_MOUSE_TIMER);
  MouseLastX = event.GetX();
  MouseLastY = event.GetY();
  TimerMouseMove->Start(100, wxTIMER_ONE_SHOT);
}

void MyMapView::OnTimerMouseMove(wxTimerEvent & WXUNUSED(event))
{
//
// MOUSE motion - Timer event handler
// 
// on-the-fly Identify
//
  OnTheFlyIdentify(MouseLastX, MouseLastY);
}

void MyMapView::OnTimerDynamic(wxTimerEvent & WXUNUSED(event))
{
//
// Dynamic Timer event handler
// 
// marking the Current Geometry
//
  ResetScreenBitmap();
  DynamicPhase++;
  if (DynamicPhase < 60)
    {
      if (!TimerDynamic)
        TimerDynamic = new wxTimer(this, ID_DYNAMIC_TIMER);
      TimerDynamic->Start(125, wxTIMER_ONE_SHOT);
  } else
    ClearMapTip();
}

void MyMapView::OnMouseWheel(wxMouseEvent & event)
{
//
// MOUSE wheel rotation - event handler
//
  int tics = event.GetWheelRotation() / event.GetWheelDelta();
  if (!TimerMouseWheel)
    TimerMouseWheel = new wxTimer(this, ID_WHEEL_TIMER);
  WheelTics += tics;
  TimerMouseWheel->Start(100, wxTIMER_ONE_SHOT);
}

void MyMapView::OnTimerMouseWheel(wxTimerEvent & WXUNUSED(event))
{
//
// MOUSE wheel rotation - Timer event handler
// 
  if (WheelTics > 0)
    {
      // zoom in
      double factor = 1.0 - ((double) WheelTics * 0.025);
      FrameExtentX *= factor;
      FrameExtentY *= factor;
  } else
    {
      // zoom out
      double factor = 1.0 - ((double) (WheelTics * -1) * 0.025);
      FrameExtentX /= factor;
      FrameExtentY /= factor;
    }
  WheelTics = 0;
  PrepareMap();
}

void MyMapView::OnContextMenu(wxMouseEvent & event)
{
//
// MOUSE Click - event handler [right button - context menu]
// 
  if (ValidMap == false)
    return;

  wxMenu *menu = new wxMenu();
  wxMenuItem *menuItem;
  wxPoint pt = event.GetPosition();
  menuItem =
    new wxMenuItem(menu, ID_Political,
                   wxT("&Political Map"), wxT(""), wxITEM_RADIO);
  menu->Append(menuItem);
  menuItem =
    new wxMenuItem(menu, ID_Satellite,
                   wxT("&Satellite Map"), wxT(""), wxITEM_RADIO);
  menu->Append(menuItem);
  menuItem =
    new wxMenuItem(menu, ID_Relief, wxT("&Relief Map"), wxT(""), wxITEM_RADIO);
  menu->Append(menuItem);
  menuItem =
    new wxMenuItem(menu, ID_Bathymetry,
                   wxT("&Bathymetry Map"), wxT(""), wxITEM_RADIO);
  menu->Append(menuItem);
  if ((MainFrame->GetTempJanHandle() != NULL
       && MainFrame->GetTempAprHandle() != NULL
       && MainFrame->GetTempJulHandle() != NULL
       && MainFrame->GetTempOctHandle() != NULL)
      || (MainFrame->GetRainJanHandle() != NULL
          && MainFrame->GetRainAprHandle() != NULL
          && MainFrame->GetRainJulHandle() != NULL
          && MainFrame->GetRainOctHandle() != NULL))
    {
      menuItem =
        new wxMenuItem(menu, ID_Temperature,
                       wxT("Mean &Temperature"), wxT(""), wxITEM_RADIO);
      menu->Append(menuItem);
      menuItem =
        new wxMenuItem(menu, ID_Rain,
                       wxT("&Precipitation (rain)"), wxT(""), wxITEM_RADIO);
      menu->Append(menuItem);
      menu->AppendSeparator();
      menuItem =
        new wxMenuItem(menu, ID_January,
                       wxT("current month: January"), wxT(""), wxITEM_RADIO);
      menu->Append(menuItem);
      menuItem =
        new wxMenuItem(menu, ID_April,
                       wxT("current month: April"), wxT(""), wxITEM_RADIO);
      menu->Append(menuItem);
      menuItem =
        new wxMenuItem(menu, ID_July,
                       wxT("current month: July"), wxT(""), wxITEM_RADIO);
      menu->Append(menuItem);
      menuItem =
        new wxMenuItem(menu, ID_October,
                       wxT("current month: October"), wxT(""), wxITEM_RADIO);
      menu->Append(menuItem);
    }
  menu->AppendSeparator();
  menuItem =
    new wxMenuItem(menu, ID_Graticule, wxT("Graticule"), wxT(""), wxITEM_CHECK);
  menu->Append(menuItem);
  menu->AppendSeparator();
  menuItem = new wxMenuItem(menu, ID_Center, wxT("Zoom to &Full Extent"));
  menuItem->SetBitmap(*(MainFrame->GetBtnCenter()));
  menu->Append(menuItem);
  menuItem =
    new wxMenuItem(menu, ID_BestZoom, wxT("Zoom to &Raster full resolution"));
  menuItem->SetBitmap(*(MainFrame->GetBtnBestZoom()));
  menu->Append(menuItem);
  menuItem =
    new wxMenuItem(menu, ID_ZoomIn, wxT("Zoom &In"), wxT(""), wxITEM_RADIO);
  menu->Append(menuItem);
  menuItem =
    new wxMenuItem(menu, ID_ZoomOut, wxT("Zoom &Out"), wxT(""), wxITEM_RADIO);
  menu->Append(menuItem);
  menuItem = new wxMenuItem(menu, ID_Pan, wxT("&Pan"), wxT(""), wxITEM_RADIO);
  menu->Append(menuItem);
  menuItem =
    new wxMenuItem(menu, ID_Identify, wxT("Identify"), wxT(""), wxITEM_RADIO);
  menu->Append(menuItem);
  menuItem =
    new wxMenuItem(menu, ID_Distance,
                   wxT("Measure Distance"), wxT(""), wxITEM_RADIO);
  menu->Append(menuItem);
// enabling/disabling items
  if (MainFrame->GetReliefHandle() == NULL)
    menu->Enable(ID_Relief, false);
  if (MainFrame->GetTempJanHandle() == NULL
      || MainFrame->GetTempAprHandle() == NULL
      || MainFrame->GetTempJulHandle() == NULL
      || MainFrame->GetTempOctHandle() == NULL)
    menu->Enable(ID_Temperature, false);
  if (MainFrame->GetRainJanHandle() == NULL
      || MainFrame->GetRainAprHandle() == NULL
      || MainFrame->GetRainJulHandle() == NULL
      || MainFrame->GetRainOctHandle() == NULL)
    menu->Enable(ID_Rain, false);
  if ((MainFrame->GetTempJanHandle() == NULL
       || MainFrame->GetTempAprHandle() == NULL
       || MainFrame->GetTempJulHandle() == NULL
       || MainFrame->GetTempOctHandle() == NULL)
      && (MainFrame->GetRainJanHandle() == NULL
          || MainFrame->GetRainAprHandle() == NULL
          || MainFrame->GetRainJulHandle() == NULL
          || MainFrame->GetRainOctHandle() == NULL))
    {
      menu->Enable(ID_January, false);
      menu->Enable(ID_April, false);
      menu->Enable(ID_July, false);
      menu->Enable(ID_October, false);
    }
  if (MainFrame->HasGraticule() == false)
    menu->Enable(ID_Graticule, false);
// setting items state
  if (MainFrame->HasCountries() == true || MainFrame->HasAdminAreas() == true)
    ;
  else
    menu->Enable(ID_Political, false);
  menu->Check(ID_Political, MainFrame->IsPoliticalMap());
  if (MainFrame->GetSatelliteHandle() == NULL)
    menu->Enable(ID_Satellite, false);
  menu->Check(ID_Satellite, MainFrame->IsSatelliteMap());
  if (MainFrame->GetReliefHandle() == NULL)
    menu->Enable(ID_Relief, false);
  menu->Check(ID_Relief, MainFrame->IsReliefMap());
  if (MainFrame->GetBathymetryHandle() == NULL)
    menu->Enable(ID_Bathymetry, false);
  menu->Check(ID_Bathymetry, MainFrame->IsBathymetryMap());
  if (MainFrame->IsTMeanJanMap() || MainFrame->IsTMeanAprMap()
      || MainFrame->IsTMeanJulMap() || MainFrame->IsTMeanOctMap())
    menu->Check(ID_Temperature, true);
  if (MainFrame->IsRainJanMap() || MainFrame->IsRainAprMap()
      || MainFrame->IsRainJulMap() || MainFrame->IsRainOctMap())
    menu->Check(ID_Rain, true);
  if (MainFrame->IsTMeanJanMap() || MainFrame->IsRainJanMap())
    menu->Check(ID_January, true);
  if (MainFrame->IsTMeanAprMap() || MainFrame->IsRainAprMap())
    menu->Check(ID_April, true);
  if (MainFrame->IsTMeanJulMap() || MainFrame->IsRainJulMap())
    menu->Check(ID_July, true);
  if (MainFrame->IsTMeanOctMap() || MainFrame->IsRainOctMap())
    menu->Check(ID_October, true);
  if (MainFrame->IsModeGraticule() == true)
    menu->Check(ID_Graticule, true);
  if (MainFrame->IsModeDistance() == true)
    menu->Check(ID_Distance, true);
  if (MainFrame->IsModeIdentify() == true)
    menu->Check(ID_Identify, true);
  if (MainFrame->IsModeZoomIn() == true)
    menu->Check(ID_ZoomIn, true);
  if (MainFrame->IsModeZoomOut() == true)
    menu->Check(ID_ZoomOut, true);
  if (MainFrame->IsModePan() == true)
    menu->Check(ID_Pan, true);
  PopupMenu(menu, pt);
}

void MyMapView::OnMouseClick(wxMouseEvent & event)
{
//
// MOUSE Click - event handler
// 
  SetFocus();

  if (MainFrame->IsModeDistance() == true)
    {
      // measured distance
      if (DistanceX1 == DBL_MAX && DistanceY1 == DBL_MAX)
        {
          // setting the first point
          DistanceX1 = (FrameMinX + (event.GetX() * PixelRatio));
          DistanceY1 = (FrameMaxY - (event.GetY() * PixelRatio));
          if (DistanceX1 > MainFrame->GetMapMaxX()
              || DistanceX1 < MainFrame->GetMapMinX()
              || DistanceY1 > MainFrame->GetMapMaxY()
              || DistanceY1 < MainFrame->GetMapMinY())
            {
              // invalid coords
              DistanceX1 = DBL_MAX;
              DistanceY1 = DBL_MAX;
              return;
            }
          if (DistancePath)
            {
              gaiaFreeGeomColl(DistancePath);
              DistancePath = NULL;
            }
          ResetScreenBitmap();
          return;
      } else if (DistanceX2 == DBL_MAX && DistanceY2 == DBL_MAX)
        {
          // setting the second point
          DistanceX2 = (FrameMinX + (event.GetX() * PixelRatio));
          DistanceY2 = (FrameMaxY - (event.GetY() * PixelRatio));
          if (DistanceX2 > MainFrame->GetMapMaxX()
              || DistanceX2 < MainFrame->GetMapMinX()
              || DistanceY2 > MainFrame->GetMapMaxY()
              || DistanceY2 < MainFrame->GetMapMinY())
            {
              // invalid coords
              DistanceX2 = DBL_MAX;
              DistanceY2 = DBL_MAX;
              return;
            }
          ResetScreenBitmap();
          return;
      } else
        {
          // resetting all
          DistanceX1 = DBL_MAX;
          DistanceY1 = DBL_MAX;
          DistanceX2 = DBL_MAX;
          DistanceY2 = DBL_MAX;
          ResetScreenBitmap();
          return;
        }
    }
  if (MainFrame->IsModeIdentify() == true)
    {
      // identify
      DoIdentify(event.GetX(), event.GetY());
      return;
    }
  if (MainFrame->IsModeZoomIn() == true)
    {
      // zoom in
      FrameCenterX = (FrameMinX + (event.GetX() * PixelRatio));
      FrameCenterY = (FrameMaxY - (event.GetY() * PixelRatio));
      FrameExtentX *= 0.5;
      FrameExtentY *= 0.5;
      PrepareMap();
      return;
    }
  if (MainFrame->IsModeZoomOut() == true)
    {
      // zoom out
      FrameCenterX = (FrameMinX + (event.GetX() * PixelRatio));
      FrameCenterY = (FrameMaxY - (event.GetY() * PixelRatio));
      FrameExtentX /= 0.5;
      FrameExtentY /= 0.5;
      PrepareMap();
      return;
    }

  if (MainFrame->IsModePan() == true)
    {
      DragStartX = event.GetX();
      DragStartY = event.GetY();
      LastDragX = event.GetX();
      LastDragY = event.GetY();
    }
}

void MyMapView::OnMouseDragStop(wxMouseEvent & event)
{
//
// MOUSE left button up - event handler
// 
  if (DragStartX < 0 || DragStartY < 0)
    return;
  LastDragX = event.GetX();
  LastDragY = event.GetY();
  if (MainFrame->IsModePan() == true)
    {
      // computing the new Map position
      double shiftX = (double) (LastDragX - DragStartX) * PixelRatio;
      double shiftY = (double) (LastDragY - DragStartY) * PixelRatio;
      DragStartX = -1;
      DragStartY = -1;
      FrameCenterX -= shiftX;
      FrameCenterY += shiftY;
      PrepareMap();
      ClearMapTip();
    }
  SetCursor(CursorCross);
  DragStartX = -1;
  DragStartY = -1;
  LastDragX = -1;
  LastDragY = -1;
}

void MyMapView::DragMap(int end_x, int end_y)
{
//
// Map dragging
//
  int width = FrameWidth;
  int height = FrameHeight;
  int baseX = 0;
  int baseY = 0;
  baseX = end_x - DragStartX;
  baseY = end_y - DragStartY;

  wxClientDC dc(this);
  dc.SetBrush(wxBrush(wxColour(192, 192, 192)));
  dc.DrawRectangle(-1, -1, width + 1, height + 1);
// copying the ScreenImage - dragged
  wxMemoryDC bmpDc(ScreenBitmap);
  int eff_width = width;
  int eff_height = height;
  int sx = 0;
  int sy = 0;
  int ix = 0;
  int iy = 0;
  if (baseX < 0)
    {
      ix = 0;
      sx = 0 - baseX;
      eff_width = width + baseX;
  } else
    {
      ix = baseX;
      eff_width = width - baseX;
    }
  if (baseY < 0)
    {
      iy = 0;
      sy = 0 - baseY;
      eff_height = height + baseY;
  } else
    {
      iy = baseY;
      eff_height = height - baseY;
    }
  dc.Blit(ix, iy, eff_width, eff_height, &bmpDc, sx, sy);
  LastDragX = end_x;
  LastDragY = end_y;
}

void MyMapView::SetBestZoom()
{
//
// setting the Frame to Raster full resolution zoom level
//
  PixelRatio = MainFrame->GetRasterMaxResolution(PixelRatio);
  FrameMinX = FrameCenterX - ((FrameWidth / 2.0) * PixelRatio);
  FrameMaxX = FrameCenterX + ((FrameWidth / 2.0) * PixelRatio);
  FrameMinY = FrameCenterY - ((FrameHeight / 2.0) * PixelRatio);
  FrameMaxY = FrameCenterY + ((FrameHeight / 2.0) * PixelRatio);
  FrameExtentX = FrameMaxX - FrameMinX;
  FrameExtentY = FrameMaxY - FrameMinY;
}

void MyMapView::SetFullExtent()
{
//
// setting the Frame to full extent
//
  FrameCenterX =
    MainFrame->GetMapMinX() +
    ((MainFrame->GetMapMaxX() - MainFrame->GetMapMinX()) / 2.0);
  FrameCenterY =
    MainFrame->GetMapMinY() +
    ((MainFrame->GetMapMaxY() - MainFrame->GetMapMinY()) / 2.0);
  FrameExtentX = MainFrame->GetMapMaxX() - MainFrame->GetMapMinX();
  FrameExtentY = MainFrame->GetMapMaxY() - MainFrame->GetMapMinY();
  FrameMinX = MainFrame->GetMapMinX();
  FrameMaxX = MainFrame->GetMapMaxX();
  FrameMinY = MainFrame->GetMapMinY();
  FrameMinY = MainFrame->GetMapMaxY();
  ValidMap = true;
}

bool MyMapView::LayerIsVisible(int layer)
{
// determining visibility scale ranges
  switch (layer)
    {
      case ATLAS_GRATICULES:
        if (MainFrame->HasGraticules() == true)
          return true;
      case ATLAS_GEOGRAPHIC_LINES:
        if (MainFrame->HasGeographicLines() == true)
          return true;
        break;
      case ATLAS_ADMIN_AREAS_1:
        if (MainFrame->HasAdminAreas() == true && CurrentScale > 1500000)
          return true;
        break;
      case ATLAS_ADMIN_AREAS_2:
        if (MainFrame->HasAdminAreas() == true && CurrentScale <= 1500000
            && CurrentScale >= 250000)
          return true;
        break;
      case ATLAS_ADMIN_AREAS_3:
        if (MainFrame->HasAdminAreas() == true && CurrentScale < 250000)
          return true;
        break;
      case ATLAS_NATURAL_RESERVES:
        if (MainFrame->HasNaturalReserves() == true)
          {
            if (MainFrame->IsWorldWideMap() == true)
              {
                if (CurrentScale < 40000000 && CurrentScale > 0)
                  return true;
            } else
              {
                if (CurrentScale < 1000000 && CurrentScale > 0)
                  return true;
              }
          }
        break;
      case ATLAS_LAKES:
        if (MainFrame->HasLakes() == true)
          {
            if (MainFrame->IsWorldWideMap() == true)
              {
                if (CurrentScale < 40000000 && CurrentScale > 0)
                  return true;
            } else
              {
                if (CurrentScale < 1000000 && CurrentScale > 0)
                  return true;
              }
          }
        break;
      case ATLAS_RIVERS:
        if (MainFrame->HasRivers() == true)
          {
            if (MainFrame->IsWorldWideMap() == true)
              {
                if (CurrentScale < 40000000 && CurrentScale > 0)
                  return true;
            } else
              {
                if (CurrentScale < 500000 && CurrentScale > 0)
                  return true;
              }
          }
        break;
      case ATLAS_RAILWAYS:
        if (MainFrame->HasRailways() == true)
          {
            if (MainFrame->IsWorldWideMap() == true)
              {
                if (CurrentScale < 40000000 && CurrentScale > 0)
                  return true;
            } else
              {
                if (CurrentScale < 4000000 && CurrentScale > 0)
                  return true;
              }
          }
        break;
      case ATLAS_RAILWAY_STATIONS:
        if (MainFrame->HasRailwayStations() == true)
          {
            if (MainFrame->IsWorldWideMap() == true)
              {
                if (CurrentScale < 40000000 && CurrentScale > 0)
                  return true;
            } else
              {
                if (CurrentScale < 1000000 && CurrentScale > 0)
                  return true;
              }
          }
        break;
      case ATLAS_CULTURAL_HERITAGE:
        if (MainFrame->HasCulturalHeritage() == true)
          {
            if (MainFrame->IsWorldWideMap() == true)
              {
                if (CurrentScale < 40000000 && CurrentScale > 0)
                  return true;
            } else
              {
                if (CurrentScale < 750000 && CurrentScale > 0)
                  return true;
              }
          }
        break;
      case ATLAS_NAMES:
        if (MainFrame->HasNames() == true)
          {
            if (MainFrame->IsWorldWideMap() == true)
              {
                if (CurrentScale < 40000000 && CurrentScale > 0)
                  return true;
            } else
              {
                if (CurrentScale < 750000 && CurrentScale > 0)
                  return true;
              }
          }
        break;
      case ATLAS_HIGHWAYS:
        if (MainFrame->HasHighways() == true)
          {
            if (MainFrame->IsWorldWideMap() == true)
              {
                if (CurrentScale < 10000000 && CurrentScale > 0)
                  return true;
            } else
              {
                if (CurrentScale < 250000 && CurrentScale > 0)
                  return true;
              }
          }
        break;
      case ATLAS_TOWNS_MINOR:
        if (MainFrame->HasTowns() == true)
          {
            if (MainFrame->IsWorldWideMap() == true)
              {
                if (CurrentScale < 5000000 && CurrentScale > 0)
                  return true;
            } else
              {
                if (CurrentScale < 500000 && CurrentScale > 0)
                  return true;
              }
          }
        break;
      case ATLAS_TOWNS_MEDIUM:
        if (MainFrame->HasTowns() == true)
          {
            if (MainFrame->IsWorldWideMap() == true)
              {
                if (CurrentScale < 10000000 && CurrentScale > 0)
                  return true;
            } else
              {
                if (CurrentScale < 1000000 && CurrentScale > 0)
                  return true;
              }
          }
        break;
      case ATLAS_TOWNS_MAJOR:
        if (MainFrame->HasTowns() == true)
          {
            if (MainFrame->IsWorldWideMap() == true)
              {
                if (CurrentScale < 60000000 && CurrentScale > 0)
                  return true;
            } else
              {
                if (CurrentScale < 6000000 && CurrentScale > 0)
                  return true;
              }
          }
        break;
      case ATLAS_TOWN_AREAS_MINOR:
        if (MainFrame->HasTownAreas() == true)
          {
            if (MainFrame->IsWorldWideMap() == true)
              {
                if (CurrentScale < 5000000 && CurrentScale > 0)
                  return true;
            } else
              {
                if (CurrentScale < 500000 && CurrentScale > 0)
                  return true;
              }
          }
        break;
      case ATLAS_TOWN_AREAS_MEDIUM:
        if (MainFrame->HasTownAreas() == true)
          {
            if (MainFrame->IsWorldWideMap() == true)
              {
                if (CurrentScale < 10000000 && CurrentScale >= 5000000)
                  return true;
            } else
              {
                if (CurrentScale < 1000000 && CurrentScale >= 500000)
                  return true;
              }
          }
        break;
      case ATLAS_TOWN_AREAS_MAJOR:
        if (MainFrame->HasTownAreas() == true)
          {
            if (MainFrame->IsWorldWideMap() == true)
              {
                if (CurrentScale < 60000000 && CurrentScale >= 10000000)
                  return true;
            } else
              {
                if (CurrentScale < 6000000 && CurrentScale >= 1000000)
                  return true;
              }
          }
        break;
    };
  return false;
}

void MyMapView::PrepareMap()
{
//
// draws the MapBitmap
//
  wxMemoryDC *dc = NULL;
  dc = new wxMemoryDC(MapBitmap);
  if (dc->IsOk() == false)
    {
      delete dc;
      return;
    }

  wxSize sz = GetClientSize();
  FrameWidth = sz.GetWidth();
  FrameHeight = sz.GetHeight();

  dc->SetBrush(wxBrush(wxColour(0, 0, 0), wxSOLID));
// initializing the bitmap background
  dc->DrawRectangle(0, 0, FrameWidth, FrameHeight);

  if (ValidMap == true)
    {
      dc->SetBrush(wxBrush(wxColour(0, 0, 0), wxSOLID));
// initializing the bitmap background
      dc->DrawRectangle(0, 0, FrameWidth, FrameHeight);
      // computing the current Frame, Ratio and Scale
      double x_ratio = FrameExtentX / (double) (FrameWidth - 4);
      double y_ratio = FrameExtentY / (double) (FrameHeight - 4);
      if (FrameExtentX / x_ratio <= FrameWidth
          && FrameExtentY / x_ratio <= FrameHeight)
        PixelRatio = x_ratio;
      else if (FrameExtentX / y_ratio <= FrameWidth
               && FrameExtentY / y_ratio <= FrameHeight)
        PixelRatio = y_ratio;
      else
        {
          if (x_ratio > y_ratio)
            PixelRatio = x_ratio;
          else
            PixelRatio = y_ratio;
        }
      FrameMinX = FrameCenterX - ((FrameWidth / 2.0) * PixelRatio);
      FrameMaxX = FrameCenterX + ((FrameWidth / 2.0) * PixelRatio);
      FrameMinY = FrameCenterY - ((FrameHeight / 2.0) * PixelRatio);
      FrameMaxY = FrameCenterY + ((FrameHeight / 2.0) * PixelRatio);
      if (FrameMinX > MainFrame->GetMapMaxX())
        {
          FrameMaxX = MainFrame->GetMapMaxX();
          FrameMinX = FrameMaxX - (FrameWidth * PixelRatio);
          FrameCenterX = (FrameMinX + ((FrameMaxX - FrameMinX) / 2.0));
      } else if (FrameMaxX < MainFrame->GetMapMinX())
        {
          FrameMinX = MainFrame->GetMapMinX();
          FrameMaxX = FrameMinX + (FrameWidth * PixelRatio);
          FrameCenterX = (FrameMinX + ((FrameMaxX - FrameMinX) / 2.0));
        }
      if (FrameMinY > MainFrame->GetMapMaxY())
        {
          FrameMaxY = MainFrame->GetMapMaxY();
          FrameMinY = FrameMaxY - (FrameHeight * PixelRatio);
          FrameCenterY = (FrameMinY + ((FrameMaxY - FrameMinY) / 2.0));
      } else if (FrameMaxY < MainFrame->GetMapMinY())
        {
          FrameMinY = MainFrame->GetMapMinY();
          FrameMaxY = FrameMinY + (FrameHeight * PixelRatio);
          FrameCenterY = (FrameMinY + ((FrameMaxY - FrameMinY) / 2.0));
        }

      if (MainFrame->IsGeographic() == true)
        {
          // current scale (Geographic Coords)
          double cm = PixelRatio * 72.0 * 0.3937;
          CurrentScale = (int) (cm * 11113200.0);
      } else
        {
          // planar coordinates
          double cm = PixelRatio * 72.0 * 0.3937;
          CurrentScale = (int) (cm * 100.0);
        }
      MainFrame->UpdateMapInfo();

      wxGraphicsContext *gr = wxGraphicsContext::Create(*dc);

      ::wxBeginBusyCursor();

      // drawing the raster map Layer (if any)
      MainFrame->SetMapInfo(ATLAS_NONE);
      if (MainFrame->IsSatelliteMap() == true)
        {
          MainFrame->SetMapInfo(ATLAS_SATELLITE);
          DrawRasterLayer(gr, ATLAS_SATELLITE);
        }
      if (MainFrame->IsReliefMap() == true)
        {
          MainFrame->SetMapInfo(ATLAS_RELIEF);
          DrawRasterLayer(gr, ATLAS_RELIEF);
        }
      if (MainFrame->IsBathymetryMap() == true)
        {
          MainFrame->SetMapInfo(ATLAS_BATHYMETRY);
          DrawRasterLayer(gr, ATLAS_BATHYMETRY);
        }
      if (MainFrame->IsTMeanJanMap() == true)
        {
          MainFrame->SetMapInfo(ATLAS_TEMPERATURE_JAN);
          DrawRasterLayer(gr, ATLAS_TEMPERATURE_JAN);
        }
      if (MainFrame->IsTMeanAprMap() == true)
        {
          MainFrame->SetMapInfo(ATLAS_TEMPERATURE_APR);
          DrawRasterLayer(gr, ATLAS_TEMPERATURE_APR);
        }
      if (MainFrame->IsTMeanJulMap() == true)
        {
          MainFrame->SetMapInfo(ATLAS_TEMPERATURE_JUL);
          DrawRasterLayer(gr, ATLAS_TEMPERATURE_JUL);
        }
      if (MainFrame->IsTMeanOctMap() == true)
        {
          MainFrame->SetMapInfo(ATLAS_TEMPERATURE_OCT);
          DrawRasterLayer(gr, ATLAS_TEMPERATURE_OCT);
        }
      if (MainFrame->IsRainJanMap() == true)
        {
          MainFrame->SetMapInfo(ATLAS_RAIN_JAN);
          DrawRasterLayer(gr, ATLAS_RAIN_JAN);
        }
      if (MainFrame->IsRainAprMap() == true)
        {
          MainFrame->SetMapInfo(ATLAS_RAIN_APR);
          DrawRasterLayer(gr, ATLAS_RAIN_APR);
        }
      if (MainFrame->IsRainJulMap() == true)
        {
          MainFrame->SetMapInfo(ATLAS_RAIN_JUL);
          DrawRasterLayer(gr, ATLAS_RAIN_JUL);
        }
      if (MainFrame->IsRainOctMap() == true)
        {
          MainFrame->SetMapInfo(ATLAS_RAIN_OCT);
          DrawRasterLayer(gr, ATLAS_RAIN_OCT);
        }
      // drawing the vector map Layers [following the zOrder]
      if (MainFrame->IsPoliticalMap() == true)
        {
          if (MainFrame->HasCountries() == true)
            DrawVectorLayer(gr, ATLAS_COUNTRIES);
          if (LayerIsVisible(ATLAS_ADMIN_AREAS_1) == true)
            DrawVectorLayer(gr, ATLAS_ADMIN_AREAS_1);
          if (LayerIsVisible(ATLAS_ADMIN_AREAS_2) == true)
            DrawVectorLayer(gr, ATLAS_ADMIN_AREAS_2);
          if (LayerIsVisible(ATLAS_ADMIN_AREAS_3) == true)
            DrawVectorLayer(gr, ATLAS_ADMIN_AREAS_3);
          if (LayerIsVisible(ATLAS_TOWN_AREAS_MINOR) == true)
            DrawVectorLayer(gr, ATLAS_TOWN_AREAS_MINOR);
          if (LayerIsVisible(ATLAS_TOWN_AREAS_MEDIUM) == true)
            DrawVectorLayer(gr, ATLAS_TOWN_AREAS_MEDIUM);
          if (LayerIsVisible(ATLAS_TOWN_AREAS_MAJOR) == true)
            DrawVectorLayer(gr, ATLAS_TOWN_AREAS_MAJOR);
          if (LayerIsVisible(ATLAS_LAKES) == true)
            DrawVectorLayer(gr, ATLAS_LAKES);
          if (LayerIsVisible(ATLAS_RIVERS) == true)
            DrawVectorLayer(gr, ATLAS_RIVERS);
          if (LayerIsVisible(ATLAS_RAILWAYS) == true)
            DrawVectorLayer(gr, ATLAS_RAILWAYS);
          if (LayerIsVisible(ATLAS_HIGHWAYS) == true)
            DrawVectorLayer(gr, ATLAS_HIGHWAYS);
      } else
        {
          if (LayerIsVisible(ATLAS_TOWN_AREAS_MINOR) == true)
            DrawVectorLayer(gr, ATLAS_TOWN_AREAS_MINOR);
          if (LayerIsVisible(ATLAS_TOWN_AREAS_MEDIUM) == true)
            DrawVectorLayer(gr, ATLAS_TOWN_AREAS_MEDIUM);
          if (LayerIsVisible(ATLAS_TOWN_AREAS_MAJOR) == true)
            DrawVectorLayer(gr, ATLAS_TOWN_AREAS_MAJOR);
          if (LayerIsVisible(ATLAS_RAILWAYS) == true)
            DrawVectorLayer(gr, ATLAS_RAILWAYS);
        }
      if (MainFrame->IsModeGraticule() == true)
        {
          if (MainFrame->HasGraticules() == true)
            DrawVectorLayer(gr, ATLAS_GRATICULES);
          if (MainFrame->HasGeographicLines() == true)
            DrawVectorLayer(gr, ATLAS_GEOGRAPHIC_LINES);
        }
      if (LayerIsVisible(ATLAS_RAILWAY_STATIONS) == true)
        DrawVectorLayer(gr, ATLAS_RAILWAY_STATIONS);
      if (LayerIsVisible(ATLAS_CULTURAL_HERITAGE) == true)
        DrawVectorLayer(gr, ATLAS_CULTURAL_HERITAGE);
      if (LayerIsVisible(ATLAS_NAMES) == true)
        DrawVectorLayer(gr, ATLAS_NAMES);
      if (LayerIsVisible(ATLAS_TOWNS_MINOR) == true)
        DrawVectorLayer(gr, ATLAS_TOWNS_MINOR);
      if (LayerIsVisible(ATLAS_TOWNS_MEDIUM) == true)
        DrawVectorLayer(gr, ATLAS_TOWNS_MEDIUM);
      if (LayerIsVisible(ATLAS_TOWNS_MAJOR) == true)
        DrawVectorLayer(gr, ATLAS_TOWNS_MAJOR);

      ::wxEndBusyCursor();

      delete gr;
    }

  dc->SetBrush(wxNullBrush);
  dc->SetPen(wxNullPen);
  delete dc;
  ResetScreenBitmap();
  Refresh();
}

void MyMapView::ResetScreenBitmap()
{
// copying the MapBitmap into the ScreenBitmap
  double great_circle;
  double geodesic;
  double length;
  bool showDistance = false;
  wxMemoryDC *mapDc = new wxMemoryDC(MapBitmap);
  wxMemoryDC *dynDc = new wxMemoryDC(DynamicMapBitmap);
  dynDc->Blit(0, 0, BitmapWidth, BitmapHeight, mapDc, 0, 0);
  delete mapDc;
  if (MainFrame->IsModeDistance() == false)
    {
      // cleaning the Distance settings
      if (DistancePath != NULL)
        gaiaFreeGeomColl(DistancePath);
      DistancePath = NULL;
      DistanceX1 = DBL_MAX;
      DistanceY1 = DBL_MAX;
      DistanceX2 = DBL_MAX;
      DistanceY2 = DBL_MAX;
    }
  if (CurrentGeometry != NULL && (DynamicPhase % 4) == 0)
    {
      wxGraphicsContext *gr = wxGraphicsContext::Create(*dynDc);
      DrawDynamicGeometry(gr, 1);
      delete gr;
    }
  if (CurrentGeometry != NULL && (DynamicPhase % 4) == 1)
    {
      wxGraphicsContext *gr = wxGraphicsContext::Create(*dynDc);
      DrawDynamicGeometry(gr, 2);
      delete gr;
    }
  if (CurrentGeometry != NULL && (DynamicPhase % 4) == 2)
    {
      wxGraphicsContext *gr = wxGraphicsContext::Create(*dynDc);
      DrawDynamicGeometry(gr, 3);
      delete gr;
    }
  if (DistanceX2 != DBL_MAX && DistanceY2 != DBL_MAX
      && DistanceX1 != DBL_MAX && DistanceY1 != DBL_MAX)
    {

      if (MainFrame->IsGeographic() == true)
        ComputeGreatCircleArc(&great_circle, &geodesic);
      else
        ComputePlanarDistance(&length);
      wxGraphicsContext *gr = wxGraphicsContext::Create(*dynDc);
      DrawDistancePath(gr);
      DrawDistanceMarker(gr, DistanceX1, DistanceY1);
      DrawDistanceMarker(gr, DistanceX2, DistanceY2);
      delete gr;
      DistanceX2 = DBL_MAX;
      DistanceY2 = DBL_MAX;
      DistanceX1 = DBL_MAX;
      DistanceY1 = DBL_MAX;
      showDistance = true;
  } else if (DistanceX1 != DBL_MAX && DistanceY1 != DBL_MAX)
    {
      wxGraphicsContext *gr = wxGraphicsContext::Create(*dynDc);
      DrawDistanceMarker(gr, DistanceX1, DistanceY1);
      delete gr;
  } else if (DistancePath != NULL)
    {
      double x1;
      double y1;
      double x2;
      double y2;
      wxGraphicsContext *gr = wxGraphicsContext::Create(*dynDc);
      DrawDistancePath(gr);
      gaiaLinestringPtr ln = DistancePath->FirstLinestring;
      while (ln)
        {
          if (ln == DistancePath->FirstLinestring)
            gaiaGetPoint(ln->Coords, 0, &x1, &y1);
          int pts = ln->Points - 1;
          gaiaGetPoint(ln->Coords, pts, &x2, &y2);
          ln = ln->Next;
        }
      DrawDistanceMarker(gr, x1, y1);
      DrawDistanceMarker(gr, x2, y2);
      delete gr;
  } else
    {
      if (DistancePath != NULL)
        gaiaFreeGeomColl(DistancePath);
      DistancePath = NULL;
    }
  wxMemoryDC *screenDc = new wxMemoryDC(ScreenBitmap);
  screenDc->Blit(0, 0, BitmapWidth, BitmapHeight, dynDc, 0, 0);
  delete dynDc;
  delete screenDc;
  Refresh();
  if (showDistance == true)
    {
      // measured distance message
      if (MainFrame->IsGeographic() == true)
        {
          wxString msg =
            wxT("Great Circle Distance: [assuming a spherical World]\n");
          char buf[128];
          sprintf(buf, "%1.2f Km\n\n", great_circle);
          msg += wxString::FromUTF8(buf);
          msg +=
            wxT("Geodesic Distance: [exactly measured on the Ellipsoid]\n");
          sprintf(buf, "%1.2f Km\n", geodesic);
          msg += wxString::FromUTF8(buf);
          wxMessageBox(msg, wxT("Measured Distance"), wxOK | wxICON_INFORMATION,
                       this);
      } else
        {
          wxString msg = wxT("Planar Distance:\n");
          char buf[128];
          sprintf(buf, "%1.2f Km\n\n", length);
          msg += wxString::FromUTF8(buf);
          wxMessageBox(msg, wxT("Measured Distance"), wxOK | wxICON_INFORMATION,
                       this);
        }
    }
}

bool MyMapView::Graticules5Degrees()
{
// testing if 5 degrees graticules are currently visible
  if (CurrentScale > 5000000 && CurrentScale <= 20000000)
    return true;
  return false;
}

bool MyMapView::Graticules10Degrees()
{
// testing if 10 degrees graticules are currently visible
  if (CurrentScale > 20000000)
    return true;
  return false;
}

void MyMapView::DrawVectorLayer(wxGraphicsContext * gr, int layer)
{
//
// drawing a Vector Layer
//
  sqlite3 *sqlite = MainFrame->GetSqlite();
  int ret;
  wxString sql;
  sqlite3_stmt *stmt;
  char xSql[1024];
  double frameMinX = FrameMinX;
  double frameMaxX = FrameMaxX;
  double frameMinY = FrameMinY;
  double frameMaxY = FrameMaxY;
  wxString table;

  switch (layer)
    {
// preparing the SQL statement
      case ATLAS_COUNTRIES:
        sql =
          wxT("SELECT geom, color FROM ") + MainFrame->GetCountries() +
          wxT("  WHERE ");
        table = MainFrame->GetCountries();
        break;
      case ATLAS_ADMIN_AREAS_1:
        sql =
          wxT("SELECT geom, color FROM ") + MainFrame->GetAdminAreas() +
          wxT(" WHERE class = '1' AND ");
        table = MainFrame->GetAdminAreas();
        break;
      case ATLAS_ADMIN_AREAS_2:
        sql =
          wxT("SELECT geom, color FROM ") + MainFrame->GetAdminAreas() +
          wxT(" WHERE class = '2' AND ");
        table = MainFrame->GetAdminAreas();
        break;
      case ATLAS_ADMIN_AREAS_3:
        sql =
          wxT("SELECT geom, color FROM ") + MainFrame->GetAdminAreas() +
          wxT(" WHERE class = '3' AND ");
        table = MainFrame->GetAdminAreas();
        break;
      case ATLAS_LAKES:
        sql = wxT("SELECT geom FROM ") + MainFrame->GetLakes() + wxT(" WHERE ");
        table = MainFrame->GetLakes();
        break;
      case ATLAS_RIVERS:
        sql =
          wxT("SELECT geom FROM ") + MainFrame->GetRivers() + wxT(" WHERE ");
        table = MainFrame->GetRivers();
        break;
      case ATLAS_RAILWAYS:
        sql =
          wxT("SELECT geom FROM ") + MainFrame->GetRailways() + wxT(" WHERE ");
        table = MainFrame->GetRailways();
        break;
      case ATLAS_HIGHWAYS:
        sql =
          wxT("SELECT geom FROM ") + MainFrame->GetHighways() + wxT(" WHERE ");
        table = MainFrame->GetHighways();
        break;
      case ATLAS_TOWNS_MINOR:
        if (MainFrame->IsWorldWideMap() == true)
          sql =
            wxT("SELECT geom FROM ") + MainFrame->GetTowns() +
            wxT(" WHERE type NOT IN ('PPLA', 'PPLC') AND ");
        else
          sql =
            wxT("SELECT geom FROM ") + MainFrame->GetTowns() +
            wxT(" WHERE population IS NULL AND ");
        table = MainFrame->GetTowns();
        break;
      case ATLAS_TOWNS_MEDIUM:
        if (MainFrame->IsWorldWideMap() == true)
          sql =
            wxT("SELECT geom FROM ") + MainFrame->GetTowns() +
            wxT(" WHERE type = 'PPLA' AND ");
        else
          sql =
            wxT("SELECT geom FROM ") + MainFrame->GetTowns() +
            wxT(" WHERE population >= 10000 AND population <= 100000 AND ");
        table = MainFrame->GetTowns();
        break;
      case ATLAS_TOWNS_MAJOR:
        if (MainFrame->IsWorldWideMap() == true)
          sql =
            wxT("SELECT geom FROM ") + MainFrame->GetTowns() +
            wxT(" WHERE type = 'PPLC' AND ");
        else
          sql =
            wxT("SELECT geom FROM ") + MainFrame->GetTowns() +
            wxT(" WHERE population > 100000 AND ");
        table = MainFrame->GetTowns();
        break;
      case ATLAS_TOWN_AREAS_MAJOR:
        if (MainFrame->IsWorldWideMap() == true)
          sql =
            wxT("SELECT geom FROM ") + MainFrame->GetTownAreas() +
            wxT(" WHERE type = 'PPLC' AND ");
        else
          sql =
            wxT("SELECT geom FROM ") + MainFrame->GetTownAreas() +
            wxT(" WHERE population > 100000 AND ");
        table = MainFrame->GetTownAreas();
        break;
      case ATLAS_TOWN_AREAS_MEDIUM:
        if (MainFrame->IsWorldWideMap() == true)
          sql =
            wxT("SELECT geom FROM ") + MainFrame->GetTownAreas() +
            wxT(" WHERE type IN ('PPLC', 'PPLA') AND ");
        else
          sql =
            wxT("SELECT geom FROM ") + MainFrame->GetTownAreas() +
            wxT(" WHERE population >= 10000 AND ");
        table = MainFrame->GetTownAreas();
        break;
      case ATLAS_TOWN_AREAS_MINOR:
        sql =
          wxT("SELECT geom FROM ") + MainFrame->GetTownAreas() + wxT(" WHERE ");
        table = MainFrame->GetTownAreas();
        break;
      case ATLAS_RAILWAY_STATIONS:
        sql =
          wxT("SELECT geom FROM ") + MainFrame->GetRailwayStations() +
          wxT(" WHERE  ");
        table = MainFrame->GetRailwayStations();
        break;
      case ATLAS_CULTURAL_HERITAGE:
        sql =
          wxT("SELECT geom FROM ") + MainFrame->GetCulturalHeritage() +
          wxT(" WHERE  ");
        table = MainFrame->GetCulturalHeritage();
        break;
      case ATLAS_NAMES:
        sql =
          wxT("SELECT geom FROM ") + MainFrame->GetNames() + wxT(" WHERE  ");
        table = MainFrame->GetNames();
        break;
      case ATLAS_GEOGRAPHIC_LINES:
        sql =
          wxT("SELECT geom FROM ") + MainFrame->GetGeographicLines() +
          wxT(" WHERE ");
        table = MainFrame->GetGeographicLines();
        break;
      case ATLAS_GRATICULES:
        if (Graticules10Degrees() == true)
          sql =
            wxT("SELECT geom FROM ") + MainFrame->GetGraticules() +
            wxT(" WHERE class = 10 AND ");
        else if (Graticules5Degrees() == true)
          sql =
            wxT("SELECT geom FROM ") + MainFrame->GetGraticules() +
            wxT(" WHERE class IS NOT NULL AND ");
        else
          sql =
            wxT("SELECT geom FROM ") + MainFrame->GetGraticules() +
            wxT(" WHERE ");
        table = MainFrame->GetGraticules();
        break;
    };

  // using the R*Tree Spatial Index
  sql += wxT("ROWID IN (SELECT pkid FROM \"idx_");
  sql += table;
  sql += wxT("_geom");
  sql += wxT("\" WHERE xmin <= ? AND xmax >= ? AND ymin <= ? AND ymax >= ?)");

  strcpy(xSql, sql.ToUTF8());
  ret = sqlite3_prepare_v2(sqlite, xSql, sql.Len(), &stmt, NULL);
  if (ret != SQLITE_OK)
    {
      wxString err = wxString::FromUTF8(sqlite3_errmsg(sqlite));
      wxMessageBox(wxT("SQL error:") + err, wxT("LibreAtlas"),
                   wxOK | wxICON_ERROR, this);
      goto stop;
    }
// query params binding
  sqlite3_reset(stmt);
  sqlite3_clear_bindings(stmt);

  sqlite3_bind_double(stmt, 1, frameMaxX);
  sqlite3_bind_double(stmt, 2, frameMinX);
  sqlite3_bind_double(stmt, 3, frameMaxY);
  sqlite3_bind_double(stmt, 4, frameMinY);

  while (1)
    {
      // scrolling the result set 
      ret = sqlite3_step(stmt);
      if (ret == SQLITE_DONE)
        break;                  // end of result set
      if (ret == SQLITE_ROW)
        {
          gaiaGeomCollPtr geom = NULL;
          if (sqlite3_column_type(stmt, 0) == SQLITE_BLOB)
            {
              // fetching Geometry
              const void *blob = sqlite3_column_blob(stmt, 0);
              int blob_size = sqlite3_column_bytes(stmt, 0);
              geom =
                gaiaFromSpatiaLiteBlobWkb((const unsigned char *) blob,
                                          blob_size);
            }

          if (geom)
            {
              // drawing the Default geometric entity
              if (layer == ATLAS_COUNTRIES)
                {
                  const char *color =
                    (const char *) sqlite3_column_text(stmt, 1);
                  DrawPolygon(gr, color[0], geom);
              } else if (layer == ATLAS_ADMIN_AREAS_1
                         || layer == ATLAS_ADMIN_AREAS_2
                         || layer == ATLAS_ADMIN_AREAS_3)
                {
                  const char *color =
                    (const char *) sqlite3_column_text(stmt, 1);
                  DrawPolygon(gr, color[0], geom);
              } else if (layer == ATLAS_LAKES)
                {
                  DrawPolygon(gr, '-', geom);
              } else if (layer == ATLAS_RIVERS)
                {
                  wxColour colour = wxColour(0, 128, 255);
                  DrawLinestring(gr, colour, geom);
              } else if (layer == ATLAS_RAILWAYS)
                {
                  wxColour colour = wxColour(128, 0, 255);
                  DrawLinestring(gr, colour, geom, true);
              } else if (layer == ATLAS_HIGHWAYS)
                {
                  wxColour colour = wxColour(255, 0, 0);
                  DrawLinestring(gr, colour, geom);
              } else if (layer == ATLAS_GEOGRAPHIC_LINES)
                {
                  wxColour colour = wxColour(255, 128, 0);
                  DrawLinestring(gr, colour, geom);
              } else if (layer == ATLAS_GRATICULES)
                {
                  wxColour colour = wxColour(255, 255, 240);
                  DrawLinestring(gr, colour, geom);
              } else if (layer == ATLAS_TOWN_AREAS_MINOR
                         || layer == ATLAS_TOWN_AREAS_MEDIUM
                         || layer == ATLAS_TOWN_AREAS_MAJOR)
                {
                  DrawPolygon(gr, '+', geom);
              } else
                DrawPoint(gr, layer, geom);
              if (geom)
                gaiaFreeGeomColl(geom);
            }
      } else
        {
          wxString err = wxString::FromUTF8(sqlite3_errmsg(sqlite));
          wxMessageBox(wxT("SQL error:") + err, wxT("LibreAtlas"),
                       wxOK | wxICON_ERROR, this);
          sqlite3_finalize(stmt);
          goto stop;
        }
    }
  sqlite3_finalize(stmt);

stop:
  return;
}

void MyMapView::DrawPoint(wxGraphicsContext * gr, int layer,
                          gaiaGeomCollPtr geom)
{
// drawing a POINT/MULTIPOINT geometry
  gaiaPointPtr point = geom->FirstPoint;
  while (point)
    {
      int dim = 1;
      int x = (int) ((point->X - FrameMinX) / PixelRatio);
      int y = FrameHeight - (int) ((point->Y - FrameMinY) / PixelRatio);

      gr->SetPen(wxPen(wxColour(0, 0, 0), 1));
      if (layer == ATLAS_CULTURAL_HERITAGE)
        {
          gr->SetBrush(wxBrush(wxColour(255, 0, 0), wxSOLID));
          dim = 4;
      } else if (layer == ATLAS_NAMES)
        {
          gr->SetBrush(wxBrush(wxColour(128, 255, 128), wxSOLID));
          dim = 2;
      } else if (layer == ATLAS_RAILWAY_STATIONS)
        {
          gr->SetBrush(wxBrush(wxColour(128, 128, 255), wxSOLID));
          dim = 3;
      } else if (layer == ATLAS_TOWNS_MAJOR)
        {
          gr->SetBrush(wxBrush(wxColour(255, 128, 0), wxSOLID));
          dim = 4;
      } else if (layer == ATLAS_TOWNS_MEDIUM)
        {
          gr->SetBrush(wxBrush(wxColour(0, 255, 0), wxSOLID));
          dim = 3;
      } else
        {
          gr->SetBrush(wxBrush(wxColour(255, 255, 0), wxSOLID));
          dim = 2;
        }
      if (layer == ATLAS_CULTURAL_HERITAGE || layer == ATLAS_RAILWAY_STATIONS)
        gr->DrawRectangle(x - dim, y - dim, dim * 2, dim * 2);
      else
        gr->DrawEllipse(x - dim, y - dim, dim * 2, dim * 2);
      point = point->Next;
    }
}

void MyMapView::DrawLinestring(wxGraphicsContext * gr, wxColour & colour,
                               gaiaGeomCollPtr geom, bool railways)
{
// drawing a LINESTRING/MULTILINESTRING geometry
  if (railways == true)
    {
      if (MainFrame->IsPoliticalMap() == true)
        gr->SetPen(wxPen(colour, 2));
      else
        gr->SetPen(wxPen(wxColour(255, 255, 255), 2));
  } else
    gr->SetPen(wxPen(colour, 1));
  gaiaLinestringPtr line = geom->FirstLinestring;
  while (line)
    {
      int iv;
      double dx;
      double dy;
      double z;
      double m;
      int x;
      int y;
      int lastX = 0;
      int lastY = 0;
      wxGraphicsPath path = gr->CreatePath();
      for (iv = 0; iv < line->Points; iv++)
        {
          if (line->DimensionModel == GAIA_XY_Z)
            {
              gaiaGetPointXYZ(line->Coords, iv, &dx, &dy, &z);
          } else if (line->DimensionModel == GAIA_XY_M)
            {
              gaiaGetPointXYM(line->Coords, iv, &dx, &dy, &m);
          } else if (line->DimensionModel == GAIA_XY_Z_M)
            {
              gaiaGetPointXYZM(line->Coords, iv, &dx, &dy, &z, &m);
          } else
            {
              gaiaGetPoint(line->Coords, iv, &dx, &dy);
            }
          x = (int) ((dx - FrameMinX) / PixelRatio);
          y = FrameHeight - (int) ((dy - FrameMinY) / PixelRatio);
          if (iv == 0)
            {
              path.MoveToPoint(x, y);
              lastX = x;
              lastY = y;
          } else
            {
              if (x == lastX && y == lastY)
                ;
              else
                {
                  path.AddLineToPoint(x, y);
                  lastX = x;
                  lastY = y;
                }
            }
        }
      gr->StrokePath(path);
      line = line->Next;
    }
}

void MyMapView::DrawPolygon(wxGraphicsContext * gr, char color,
                            gaiaGeomCollPtr geom)
{
// drawing a POLYGON/MULTIPOLYGON geometry
  gr->SetPen(wxPen(wxColour(0, 0, 0), 1));
  if (color == '-')
    gr->SetBrush(wxBrush(wxColour(0, 128, 255), wxSOLID));
  else if (color == '+')
    {
      if (MainFrame->IsPoliticalMap() == true)
        gr->SetBrush(wxBrush(wxColour(255, 128, 0, 128), wxSOLID));
      else
        gr->SetBrush(wxBrush(wxColour(255, 255, 0, 32), wxSOLID));
  } else if (color == 'R')
    gr->SetBrush(wxBrush(wxColour(255, 192, 192), wxSOLID));
  else if (color == 'G')
    gr->SetBrush(wxBrush(wxColour(192, 232, 192), wxSOLID));
  else if (color == 'B')
    gr->SetBrush(wxBrush(wxColour(192, 192, 255), wxSOLID));
  else if (color == 'M')
    gr->SetBrush(wxBrush(wxColour(255, 192, 255), wxSOLID));
  else if (color == 'C')
    gr->SetBrush(wxBrush(wxColour(192, 255, 255), wxSOLID));
  else if (color == 'W')
    gr->SetBrush(wxBrush(wxColour(255, 255, 255), wxSOLID));
  else
    gr->SetBrush(wxBrush(wxColour(232, 232, 192), wxSOLID));

  gaiaPolygonPtr polyg = geom->FirstPolygon;
  while (polyg)
    {
      int iv;
      double dx;
      double dy;
      double z;
      double m;
      int x;
      int y;
      int lastX = 0;
      int lastY = 0;
      int ib;
      wxGraphicsPath path = gr->CreatePath();
      gaiaRingPtr ring = polyg->Exterior;
      // exterior border
      for (iv = 0; iv < ring->Points; iv++)
        {
          if (ring->DimensionModel == GAIA_XY_Z)
            {
              gaiaGetPointXYZ(ring->Coords, iv, &dx, &dy, &z);
          } else if (ring->DimensionModel == GAIA_XY_M)
            {
              gaiaGetPointXYM(ring->Coords, iv, &dx, &dy, &m);
          } else if (ring->DimensionModel == GAIA_XY_Z_M)
            {
              gaiaGetPointXYZM(ring->Coords, iv, &dx, &dy, &z, &m);
          } else
            {
              gaiaGetPoint(ring->Coords, iv, &dx, &dy);
            }
          x = (int) ((dx - FrameMinX) / PixelRatio);
          y = FrameHeight - (int) ((dy - FrameMinY) / PixelRatio);
          if (iv == 0)
            {
              path.MoveToPoint(x, y);
              lastX = x;
              lastY = y;
          } else
            {
              if (x == lastX && y == lastY)
                ;
              else
                {
                  path.AddLineToPoint(x, y);
                  lastX = x;
                  lastY = y;
                }
            }
        }
      path.CloseSubpath();
      for (ib = 0; ib < polyg->NumInteriors; ib++)
        {
          // interior borders
          ring = polyg->Interiors + ib;
          for (iv = 0; iv < ring->Points; iv++)
            {
              if (ring->DimensionModel == GAIA_XY_Z)
                {
                  gaiaGetPointXYZ(ring->Coords, iv, &dx, &dy, &z);
              } else if (ring->DimensionModel == GAIA_XY_M)
                {
                  gaiaGetPointXYM(ring->Coords, iv, &dx, &dy, &m);
              } else if (ring->DimensionModel == GAIA_XY_Z_M)
                {
                  gaiaGetPointXYZM(ring->Coords, iv, &dx, &dy, &z, &m);
              } else
                {
                  gaiaGetPoint(ring->Coords, iv, &dx, &dy);
                }
              x = (int) ((dx - FrameMinX) / PixelRatio);
              y = FrameHeight - (int) ((dy - FrameMinY) / PixelRatio);
              if (iv == 0)
                {
                  path.MoveToPoint(x, y);
                  lastX = x;
                  lastY = y;
              } else
                {
                  if (x == lastX && y == lastY)
                    ;
                  else
                    {
                      path.AddLineToPoint(x, y);
                      lastX = x;
                      lastY = y;
                    }
                }
            }
          path.CloseSubpath();
        }
      gr->FillPath(path);
      gr->StrokePath(path);
      polyg = polyg->Next;
    }
}

void MyMapView::DrawDistanceMarker(wxGraphicsContext * gr, double xx, double yy)
{
// drawing a Distance marker
  int x = (int) ((xx - FrameMinX) / PixelRatio);
  int y = FrameHeight - (int) ((yy - FrameMinY) / PixelRatio);
  gr->DrawBitmap(Pin, x - 1, y - 15, Pin.GetWidth(), Pin.GetHeight());
}

bool MyMapView::ComputeGreatCircleLength(double *great_circle, double *geodesic)
{
// computing the Great Circle distance
  bool ok = false;
  int ret;
  sqlite3_stmt *stmt;
  sqlite3 *sqlite = MainFrame->GetSqlite();
  if (!sqlite)
    return false;

  wxString sql =
    wxT("SELECT CvtToKm(GreatCircleLength(g)), CvtToKm(GeodesicLength(g)) ");
  sql += wxT("FROM (SELECT ST_GeomFromText('LINESTRING('");
  sql += wxT("|| ? || ' ' || ? || ', ' || ? || ' ' || ? || ')', ?) AS g)");
  ret = sqlite3_prepare_v2(sqlite, sql.ToUTF8(), sql.Len(), &stmt, NULL);
  if (ret != SQLITE_OK)
    {
      wxString err = wxString::FromUTF8(sqlite3_errmsg(sqlite));
      wxMessageBox(wxT("SQL error:") + err, wxT("LibreAtlas"),
                   wxOK | wxICON_ERROR, this);
      return false;
    }
// query params binding
  sqlite3_reset(stmt);
  sqlite3_clear_bindings(stmt);
  sqlite3_bind_double(stmt, 1, DistanceX1);
  sqlite3_bind_double(stmt, 2, DistanceY1);
  sqlite3_bind_double(stmt, 3, DistanceX2);
  sqlite3_bind_double(stmt, 4, DistanceY2);
  sqlite3_bind_int(stmt, 5, MainFrame->GetSrid());

  while (1)
    {
      // scrolling the result set 
      ret = sqlite3_step(stmt);
      if (ret == SQLITE_DONE)
        break;                  // end of result set
      if (ret == SQLITE_ROW)
        {
          if (sqlite3_column_type(stmt, 0) == SQLITE_FLOAT
              && sqlite3_column_type(stmt, 1) == SQLITE_FLOAT)
            {
              *great_circle = sqlite3_column_double(stmt, 0);
              *geodesic = sqlite3_column_double(stmt, 1);
              ok = true;
            }
      } else
        {
          wxString err = wxString::FromUTF8(sqlite3_errmsg(sqlite));
          wxMessageBox(wxT("SQL error:") + err, wxT("LibreAtlas"),
                       wxOK | wxICON_ERROR, this);
          sqlite3_finalize(stmt);
          return false;
        }
    }
  sqlite3_finalize(stmt);
  return ok;
}

bool MyMapView::ComputePlanarLength(double *length)
{
// computing the Great Circle distance
  bool ok = false;
  int ret;
  sqlite3_stmt *stmt;
  sqlite3 *sqlite = MainFrame->GetSqlite();
  if (!sqlite)
    return false;

  wxString sql =
    wxT("SELECT CvtToKm(ST_Distance(MakePoint(?, ?), MakePoint(?, ?)))");
  ret = sqlite3_prepare_v2(sqlite, sql.ToUTF8(), sql.Len(), &stmt, NULL);
  if (ret != SQLITE_OK)
    {
      wxString err = wxString::FromUTF8(sqlite3_errmsg(sqlite));
      wxMessageBox(wxT("SQL error:") + err, wxT("LibreAtlas"),
                   wxOK | wxICON_ERROR, this);
      return false;
    }
// query params binding
  sqlite3_reset(stmt);
  sqlite3_clear_bindings(stmt);
  sqlite3_bind_double(stmt, 1, DistanceX1);
  sqlite3_bind_double(stmt, 2, DistanceY1);
  sqlite3_bind_double(stmt, 3, DistanceX2);
  sqlite3_bind_double(stmt, 4, DistanceY2);

  while (1)
    {
      // scrolling the result set 
      ret = sqlite3_step(stmt);
      if (ret == SQLITE_DONE)
        break;                  // end of result set
      if (ret == SQLITE_ROW)
        {
          if (sqlite3_column_type(stmt, 0) == SQLITE_FLOAT)
            {
              *length = sqlite3_column_double(stmt, 0);
              ok = true;
            }
      } else
        {
          wxString err = wxString::FromUTF8(sqlite3_errmsg(sqlite));
          wxMessageBox(wxT("SQL error:") + err, wxT("LibreAtlas"),
                       wxOK | wxICON_ERROR, this);
          sqlite3_finalize(stmt);
          return false;
        }
    }
  sqlite3_finalize(stmt);
  return ok;
}

void MyMapView::ComputeGreatCircleArc(double *great_circle, double *geodesic)
{
// computing a Great Circle Arc
  *great_circle = 0.0;
  *geodesic = 0.0;
  if (DistancePath != NULL)
    {
      gaiaFreeGeomColl(DistancePath);
      DistancePath = NULL;
    }

  if (DistanceX1 == DBL_MAX || DistanceY1 == DBL_MAX
      || DistanceX2 == DBL_MAX || DistanceY2 == DBL_MAX)
    return;

  GreatCirclePath arc =
    GreatCirclePath(DistanceX1, DistanceY1, DistanceX2, DistanceY2);
  DistancePath = gaiaAllocGeomColl();
  if (arc.DoesCrossesDateLine() == false)
    {
      // single line: doesn't crosses the International Date Line
      gaiaLinestringPtr ln =
        gaiaAddLinestringToGeomColl(DistancePath, arc.GetNumPoints());
      for (int iv = 0; iv < arc.GetNumPoints(); iv++)
        {
          // setting points
          GreatCirclePoint *pt = arc.GetPoint(iv);
          gaiaSetPoint(ln->Coords, iv, pt->GetX(), pt->GetY());
        }
  } else
    {
      // double line: crosses the International Date Line
      int iv;
      double side;
      gaiaLinestringPtr ln =
        gaiaAddLinestringToGeomColl(DistancePath, arc.GetNumPointsOne() + 1);
      for (iv = 0; iv < arc.GetNumPointsOne(); iv++)
        {
          // setting points
          GreatCirclePoint *pt = arc.GetPoint(iv);
          if (pt->GetX() > 0.0)
            side = 179.99;
          else
            side = -179.99;
          gaiaSetPoint(ln->Coords, iv, pt->GetX(), pt->GetY());
        }
      gaiaSetPoint(ln->Coords, iv, side, arc.GetDateLineY());

      side *= -1.0;
      ln = gaiaAddLinestringToGeomColl(DistancePath, arc.GetNumPointsTwo());
      gaiaSetPoint(ln->Coords, 0, side, arc.GetDateLineY());
      iv = 1;
      for (int i = arc.GetNumPointsOne(); i < arc.GetNumPoints(); i++)
        {
          // setting points
          GreatCirclePoint *pt = arc.GetPoint(i);
          gaiaSetPoint(ln->Coords, iv, pt->GetX(), pt->GetY());
          iv++;
        }
    }

  if (ComputeGreatCircleLength(great_circle, geodesic) != true)
    {
      *great_circle = 0.0;
      *geodesic = 0.0;
    }
}

void MyMapView::ComputePlanarDistance(double *length)
{
// computing a planar distance
  *length = 0.0;
  if (DistancePath != NULL)
    {
      gaiaFreeGeomColl(DistancePath);
      DistancePath = NULL;
    }

  if (DistanceX1 == DBL_MAX || DistanceY1 == DBL_MAX
      || DistanceX2 == DBL_MAX || DistanceY2 == DBL_MAX)
    return;

  DistancePath = gaiaAllocGeomColl();
  gaiaLinestringPtr ln = gaiaAddLinestringToGeomColl(DistancePath, 2);
  gaiaSetPoint(ln->Coords, 0, DistanceX1, DistanceY1);
  gaiaSetPoint(ln->Coords, 1, DistanceX2, DistanceY2);

  if (ComputePlanarLength(length) != true)
    *length = 0.0;
}

void MyMapView::DrawDistancePath(wxGraphicsContext * gr)
{
// drawing the Distance Path
  gaiaLinestringPtr line;

  if (DistancePath == NULL)
    return;

  gr->SetPen(wxPen(wxColour(255, 255, 0), 8));
  line = DistancePath->FirstLinestring;
  while (line)
    {
      // marking a Linestring
      int iv;
      double dx;
      double dy;
      double z;
      double m;
      int x;
      int y;
      int lastX = 0;
      int lastY = 0;
      wxGraphicsPath path = gr->CreatePath();
      for (iv = 0; iv < line->Points; iv++)
        {
          if (line->DimensionModel == GAIA_XY_Z)
            {
              gaiaGetPointXYZ(line->Coords, iv, &dx, &dy, &z);
          } else if (line->DimensionModel == GAIA_XY_M)
            {
              gaiaGetPointXYM(line->Coords, iv, &dx, &dy, &m);
          } else if (line->DimensionModel == GAIA_XY_Z_M)
            {
              gaiaGetPointXYZM(line->Coords, iv, &dx, &dy, &z, &m);
          } else
            {
              gaiaGetPoint(line->Coords, iv, &dx, &dy);
            }
          x = (int) ((dx - FrameMinX) / PixelRatio);
          y = FrameHeight - (int) ((dy - FrameMinY) / PixelRatio);
          if (iv == 0)
            {
              path.MoveToPoint(x, y);
              lastX = x;
              lastY = y;
          } else
            {
              if (x == lastX && y == lastY)
                ;
              else
                {
                  path.AddLineToPoint(x, y);
                  lastX = x;
                  lastY = y;
                }
            }
        }
      gr->StrokePath(path);
      line = line->Next;
    }

  gr->SetPen(wxPen(wxColour(0, 0, 255), 4));
  line = DistancePath->FirstLinestring;
  while (line)
    {
      // marking a Linestring
      int iv;
      double dx;
      double dy;
      double z;
      double m;
      int x;
      int y;
      int lastX = 0;
      int lastY = 0;
      wxGraphicsPath path = gr->CreatePath();
      for (iv = 0; iv < line->Points; iv++)
        {
          if (line->DimensionModel == GAIA_XY_Z)
            {
              gaiaGetPointXYZ(line->Coords, iv, &dx, &dy, &z);
          } else if (line->DimensionModel == GAIA_XY_M)
            {
              gaiaGetPointXYM(line->Coords, iv, &dx, &dy, &m);
          } else if (line->DimensionModel == GAIA_XY_Z_M)
            {
              gaiaGetPointXYZM(line->Coords, iv, &dx, &dy, &z, &m);
          } else
            {
              gaiaGetPoint(line->Coords, iv, &dx, &dy);
            }
          x = (int) ((dx - FrameMinX) / PixelRatio);
          y = FrameHeight - (int) ((dy - FrameMinY) / PixelRatio);
          if (iv == 0)
            {
              path.MoveToPoint(x, y);
              lastX = x;
              lastY = y;
          } else
            {
              if (x == lastX && y == lastY)
                ;
              else
                {
                  path.AddLineToPoint(x, y);
                  lastX = x;
                  lastY = y;
                }
            }
        }
      gr->StrokePath(path);
      line = line->Next;
    }
}

void MyMapView::DrawDynamicGeometry(wxGraphicsContext * gr, int thick)
{
// drawing the currently selected Geometry
  gaiaPointPtr point;
  gaiaLinestringPtr line;
  gaiaPolygonPtr polyg;
  gr->SetPen(wxPen(wxColour(255, 0, 0), thick));

  point = CurrentGeometry->FirstPoint;
  while (point)
    {
      // marking a Point
      int x = (int) ((point->X - FrameMinX) / PixelRatio);
      int y = FrameHeight - (int) ((point->Y - FrameMinY) / PixelRatio);
      gr->DrawEllipse(x - 6, y - 6, 12, 12);
      point = point->Next;
    }

  line = CurrentGeometry->FirstLinestring;
  while (line)
    {
      // marking a Linestring
      int iv;
      double dx;
      double dy;
      double z;
      double m;
      int x;
      int y;
      int lastX = 0;
      int lastY = 0;
      wxGraphicsPath path = gr->CreatePath();
      for (iv = 0; iv < line->Points; iv++)
        {
          if (line->DimensionModel == GAIA_XY_Z)
            {
              gaiaGetPointXYZ(line->Coords, iv, &dx, &dy, &z);
          } else if (line->DimensionModel == GAIA_XY_M)
            {
              gaiaGetPointXYM(line->Coords, iv, &dx, &dy, &m);
          } else if (line->DimensionModel == GAIA_XY_Z_M)
            {
              gaiaGetPointXYZM(line->Coords, iv, &dx, &dy, &z, &m);
          } else
            {
              gaiaGetPoint(line->Coords, iv, &dx, &dy);
            }
          x = (int) ((dx - FrameMinX) / PixelRatio);
          y = FrameHeight - (int) ((dy - FrameMinY) / PixelRatio);
          if (iv == 0)
            {
              path.MoveToPoint(x, y);
              lastX = x;
              lastY = y;
          } else
            {
              if (x == lastX && y == lastY)
                ;
              else
                {
                  path.AddLineToPoint(x, y);
                  lastX = x;
                  lastY = y;
                }
            }
        }
      gr->StrokePath(path);
      line = line->Next;
    }

  polyg = CurrentGeometry->FirstPolygon;
  while (polyg)
    {
      // marking Polygon borders
      int iv;
      double dx;
      double dy;
      double z;
      double m;
      int x;
      int y;
      int lastX = 0;
      int lastY = 0;
      int ib;
      wxGraphicsPath path = gr->CreatePath();
      gaiaRingPtr ring = polyg->Exterior;
      // exterior border
      for (iv = 0; iv < ring->Points; iv++)
        {
          if (ring->DimensionModel == GAIA_XY_Z)
            {
              gaiaGetPointXYZ(ring->Coords, iv, &dx, &dy, &z);
          } else if (ring->DimensionModel == GAIA_XY_M)
            {
              gaiaGetPointXYM(ring->Coords, iv, &dx, &dy, &m);
          } else if (ring->DimensionModel == GAIA_XY_Z_M)
            {
              gaiaGetPointXYZM(ring->Coords, iv, &dx, &dy, &z, &m);
          } else
            {
              gaiaGetPoint(ring->Coords, iv, &dx, &dy);
            }
          x = (int) ((dx - FrameMinX) / PixelRatio);
          y = FrameHeight - (int) ((dy - FrameMinY) / PixelRatio);
          if (iv == 0)
            {
              path.MoveToPoint(x, y);
              lastX = x;
              lastY = y;
          } else
            {
              if (x == lastX && y == lastY)
                ;
              else
                {
                  path.AddLineToPoint(x, y);
                  lastX = x;
                  lastY = y;
                }
            }
        }
      path.CloseSubpath();
      for (ib = 0; ib < polyg->NumInteriors; ib++)
        {
          // interior borders
          ring = polyg->Interiors + ib;
          for (iv = 0; iv < ring->Points; iv++)
            {
              if (ring->DimensionModel == GAIA_XY_Z)
                {
                  gaiaGetPointXYZ(ring->Coords, iv, &dx, &dy, &z);
              } else if (ring->DimensionModel == GAIA_XY_M)
                {
                  gaiaGetPointXYM(ring->Coords, iv, &dx, &dy, &m);
              } else if (ring->DimensionModel == GAIA_XY_Z_M)
                {
                  gaiaGetPointXYZM(ring->Coords, iv, &dx, &dy, &z, &m);
              } else
                {
                  gaiaGetPoint(ring->Coords, iv, &dx, &dy);
                }
              x = (int) ((dx - FrameMinX) / PixelRatio);
              y = FrameHeight - (int) ((dy - FrameMinY) / PixelRatio);
              if (iv == 0)
                {
                  path.MoveToPoint(x, y);
                  lastX = x;
                  lastY = y;
              } else
                {
                  if (x == lastX && y == lastY)
                    ;
                  else
                    {
                      path.AddLineToPoint(x, y);
                      lastX = x;
                      lastY = y;
                    }
                }
            }
          path.CloseSubpath();
        }
      gr->StrokePath(path);
      polyg = polyg->Next;
    }
}

void MyMapView::DrawRasterLayer(wxGraphicsContext * gr, int layer)
{
//
// drawing a Raster Layer
//
  void *handle = NULL;
  if (layer == ATLAS_SATELLITE)
    handle = MainFrame->GetSatelliteHandle();
  if (layer == ATLAS_RELIEF)
    handle = MainFrame->GetReliefHandle();
  if (layer == ATLAS_BATHYMETRY)
    handle = MainFrame->GetBathymetryHandle();
  if (layer == ATLAS_TEMPERATURE_JAN)
    handle = MainFrame->GetTempJanHandle();
  if (layer == ATLAS_TEMPERATURE_APR)
    handle = MainFrame->GetTempAprHandle();
  if (layer == ATLAS_TEMPERATURE_JUL)
    handle = MainFrame->GetTempJulHandle();
  if (layer == ATLAS_TEMPERATURE_OCT)
    handle = MainFrame->GetTempOctHandle();
  if (layer == ATLAS_RAIN_JAN)
    handle = MainFrame->GetRainJanHandle();
  if (layer == ATLAS_RAIN_APR)
    handle = MainFrame->GetRainAprHandle();
  if (layer == ATLAS_RAIN_JUL)
    handle = MainFrame->GetRainJulHandle();
  if (layer == ATLAS_RAIN_OCT)
    handle = MainFrame->GetRainOctHandle();
  if (!handle)
    return;

// trying to get the raster background image from the RasterLite data source
  wxBitmap bmp(wxImage(FrameWidth, FrameHeight));
  if (GetRaster(handle, &bmp) == false)
    {
      // error: marking a Magenta rectangle
      gr->SetBrush(wxBrush(wxColour(255, 0, 192)));
      gr->SetPen(wxPen(wxColour(192, 64, 192)));
      gr->DrawRectangle(0, 0, FrameWidth, FrameHeight);
      return;
    }
// drawing the raster background
  gr->DrawBitmap(bmp, 0, 0, FrameWidth, FrameHeight);
}

bool MyMapView::GetRaster(void *handle, wxBitmap * bmp)
{
//
// trying to build the required raster image from the RasterLite Data Source
//
  int ret;
  double pixel_x_size;
  double pixel_y_size;
  int use_rtree;
  sqlite3_stmt *stmt = NULL;
  double map_frame_width = (double) FrameWidth * PixelRatio;
  double map_frame_height = (double) FrameHeight * PixelRatio;
  double min_x = FrameCenterX - (map_frame_width / 2.0);
  double max_x = FrameCenterX + (map_frame_width / 2.0);
  double min_y = FrameCenterY - (map_frame_height / 2.0);
  double max_y = FrameCenterY + (map_frame_height / 2.0);
  if (handle == NULL)
    {
      fprintf(stderr, "invalid datasource\n");
      fflush(stderr);
      return false;
    }
  if (FrameWidth < 64 || FrameWidth > 32768 || FrameHeight < 64
      || FrameHeight > 32768)
    {
      fprintf(stderr, "invalid raster dims [%dh X %dv]\n", FrameWidth,
              FrameHeight);
      fflush(stderr);
      return false;
    }
  if (rasterliteGetBestAccess
      (handle, PixelRatio, &pixel_x_size, &pixel_y_size, &stmt,
       &use_rtree) != RASTERLITE_OK)
    {
      fprintf(stderr, "invalid Raster access method\n");
      fflush(stderr);
      return false;
    }
  wxMemoryDC *dc = NULL;
  dc = new wxMemoryDC(*bmp);
  if (dc->IsOk() == false)
    {
      delete dc;
      return false;
    }
// setting a white background
  dc->SetBrush(wxBrush(wxColour(0, 0, 0)));
  dc->SetPen(wxPen(wxColour(0, 0, 0)));
  dc->DrawRectangle(0, 0, FrameWidth, FrameHeight);
// binding query params 
  sqlite3_reset(stmt);
  sqlite3_clear_bindings(stmt);
  if (use_rtree)
    {
      // using the Spatial Index R*Tree 
      sqlite3_bind_double(stmt, 1, max_x);
      sqlite3_bind_double(stmt, 2, min_x);
      sqlite3_bind_double(stmt, 3, max_y);
      sqlite3_bind_double(stmt, 4, min_y);
  } else
    {
      // plain Table Scan 
      sqlite3_bind_double(stmt, 1, min_x);
      sqlite3_bind_double(stmt, 2, min_y);
      sqlite3_bind_double(stmt, 3, max_x);
      sqlite3_bind_double(stmt, 4, max_y);
    }
  sqlite3_bind_double(stmt, 5, pixel_x_size);
  sqlite3_bind_double(stmt, 6, pixel_y_size);
  while (1)
    {
      // scrolling the result set 
      ret = sqlite3_step(stmt);
      if (ret == SQLITE_DONE)
        break;                  // end of result set
      if (ret == SQLITE_ROW)
        {
          // retrieving query values 
          gaiaGeomCollPtr geom = NULL;
          wxImage *img = NULL;
          if (sqlite3_column_type(stmt, 0) == SQLITE_BLOB)
            {
              // fetching Geometry 
              const void *blob = sqlite3_column_blob(stmt, 0);
              int blob_size = sqlite3_column_bytes(stmt, 0);
              geom =
                gaiaFromSpatiaLiteBlobWkb((const unsigned char *)
                                          blob, blob_size);
            }
          if (sqlite3_column_type(stmt, 1) == SQLITE_BLOB)
            {
              // fetching Raster Image 
              const unsigned char *blob =
                (const unsigned char *) sqlite3_column_blob(stmt, 1);
              int blob_size = sqlite3_column_bytes(stmt, 1);
              int type = gaiaGuessBlobType(blob, blob_size);
              if (type == GAIA_JPEG_BLOB || type == GAIA_EXIF_BLOB
                  || type == GAIA_EXIF_GPS_BLOB || type == GAIA_PNG_BLOB
                  || type == GAIA_GIF_BLOB || type == GAIA_TIFF_BLOB)
                {
                  wxMemoryInputStream reader(blob, blob_size);
                  img = new wxImage(reader);
                }
            }
          if (geom && img)
            {
              // resizing the image [tile] 
              double pre_FrameWidth =
                InternalRound(((double) (img->GetWidth()) * pixel_x_size) /
                              PixelRatio);
              double pre_FrameHeight =
                InternalRound(((double) (img->GetHeight()) * pixel_y_size) /
                              PixelRatio);
              int new_FrameWidth = (int) pre_FrameWidth + 1;
              int new_FrameHeight = (int) pre_FrameHeight + 1;
              double x = (geom->MinX - min_x) / PixelRatio;
              double y =
                (double) FrameHeight - ((geom->MaxY - min_y) / PixelRatio);
              if (new_FrameWidth > (FrameWidth * 8)
                  || new_FrameHeight > (FrameHeight * 8))
                {
                  // TOO BIG: drawing a gray rectangle 
                  dc->SetBrush(wxBrush(wxColour(240, 240, 240)));
                  dc->SetPen(wxPen(wxColour(192, 192, 192)));
                  dc->DrawRectangle(IntRound(x), IntRound(y), new_FrameWidth,
                                    new_FrameHeight);
              } else
                {
                  // resizing the raster tile 
                  wxImage img2(new_FrameWidth, new_FrameHeight);
                  if (new_FrameWidth == FrameWidth
                      && new_FrameHeight == FrameHeight)
                    img2 = img->Copy();
                  else
                    img2 =
                      img->Rescale(new_FrameWidth, new_FrameHeight,
                                   wxIMAGE_QUALITY_NORMAL);
                  // drawing the raster tile
                  wxBitmap bitmap = wxBitmap(img2);
                  wxMask *mask = new wxMask(bitmap, wxColour(0, 0, 0));
                  bitmap.SetMask(mask);
                  dc->DrawBitmap(bitmap, (wxCoord) IntRound(x),
                                 (wxCoord) IntRound(y), true);
                }
            }
          if (geom)
            gaiaFreeGeomColl(geom);
          if (img)
            delete img;
      } else
        {
          fprintf(stderr, "Raster: SQL error\n");
          fflush(stderr);
          delete dc;
          return false;
        }
    }
  delete dc;
  return true;
}

double MyMapView::InternalRound(double value)
{
// replacing the C99 round() function 
  double min = floor(value);
  if (fabs(value - min) < 0.5)
    return min;
  return min + 1.0;
}

int MyMapView::IntRound(double value)
{
/* replacing the C99 round() function */
  double min = floor(value);
  if (fabs(value - min) < 0.5)
    return (int) min;
  return (int) (min + 1.0);
}
